# Copyright (c) 2012 Cloudera, Inc. All rights reserved.
import asyncore
import errno
import getopt
import logging
import os
import sys

from supervisor.childutils import EventListenerProtocol

LOG = None

class Event(object):
  """
  Helper class used to parse an event. Since we use non-blocking I/O in the
  agent case, this class helps keep things in sync across several read() calls
  on the underlying fd. It's more complex than necessary for the blocking case
  used in the event listener, but at least the code is shared.
  """

  def __init__(self):
    self.properties = { }
    self.payload = None
    self._header = ''
    self._payload_len = None

  def feed(self, data):
    """
    Feed data into the event. If the input data has more data than this
    event expects, returns the leftover data.

    @param data: The data to feed.
    @return Data that is not part of this event.
    """
    if self._payload_len is None:
      # Haven't read the whole header yet.
      data = data.split('\n', 1)
      self._header += data[0]
      if len(data) == 2:
        # Header is finished, parse all properties, and grab the
        # payload length.
        data = data[1]
        for val in self._header.split(' '):
          val = val.split(':', 1)
          if len(val) == 2:
            self.properties[val[0]] = val[1]

        self._payload_len = int(self.properties.get('len', 0))
        if self._payload_len:
          self.payload = ''
      else:
        # Header not finished, all data was consumed.
        return ''

    # Consume as much data as needed.
    needed = self.needed_data()
    if not needed:
      return data
    if needed >= len(data):
      self.payload += data
      return ''
    else:
      self.payload += data[:needed]
      return data[needed:]

  def needed_data(self):
    """
    Returns how many bytes the event still expects, or None if the header
    hasn't been read yet.
    """
    if self._payload_len:
      return self._payload_len - len(self.payload)
    return self._payload_len


class SupervisordEventMonitor(asyncore.file_dispatcher):
  """
  An asyncore dispatcher that monitors a given file for supervisord events.
  The dispatcher will be notified of read events, and code using this
  monitor can use the consume_events() method to fetch events that have
  been read from supervisord.
  """

  def __init__(self, path):
    self.heartbeat = False

    self.currEvent = Event()
    self.events = [ ]

    # For testing purposes.
    self.bytes_read = 0

    self.fifo = os.open(path, os.O_RDONLY | os.O_NONBLOCK)
    asyncore.file_dispatcher.__init__(self, self.fifo)

  def writable(self):
    return False

  def handle_read(self):
    # Read as much data as we can from the file before we're told that
    # the buffer is drained.
    buf = ''
    while True:
      try:
        buf += os.read(self.fifo, 4096)
      except OSError, ose:
        if ose.errno == errno.EWOULDBLOCK:
          break
        elif ose.errno == errno.EINTR:
          continue
        raise

    self.bytes_read += len(buf)
    while buf:
      buf = self.currEvent.feed(buf)
      if self.currEvent.needed_data() == 0:
        # Event is complete, queue it and create a new one.
        LOG.debug('Event ready: %r', self.currEvent.properties)
        self.events.append(self.currEvent)
        self.currEvent = Event()

  def consume_events(self):
    """
    Returns the list of ready events. The internal list of events is emptied
    when this method is called.
    """
    ret = self.events
    self.events = []
    return ret


if __name__ == '__main__':
  """
  The supervisord event listener implementation. Reads events from stdin
  and pipes them to the file given in the command line. Keeps supervisord
  happy by printing the right things to stdout.
  """
  opts, args = getopt.getopt(sys.argv[1:], 'l:')

  logfile = None
  for o, v in opts:
    if o == '-l':
      logfile = v

  if len(args) != 1:
    sys.exit(1)

  if not os.path.exists(args[0]):
    sys.exit(0)

  if logfile:
    logging.basicConfig(filename = logfile)
  LOG = logging.getLogger(__name__)

  supervisor = EventListenerProtocol()
  fifo = os.open(args[0], os.O_WRONLY | os.O_SYNC)
  try:
    try:
      eof = False
      while not eof:
        supervisor.ready()

        # Read the event.
        evt = Event()
        header = sys.stdin.readline()
        if not header:
          eof = True
          break

        evt.feed(header)
        if evt.needed_data() > 0:
          evt.feed(sys.stdin.read(evt.needed_data()))

        # Write the data we just read to the pipe.
        os.write(fifo, evt._header)
        os.write(fifo, '\n')
        if evt.payload:
          os.write(fifo, evt.payload)

        supervisor.ok()
    except:
      LOG.exception('Error in supervisord listener loop.')
      raise
  finally:
    os.close(fifo)
else:
  LOG = logging.getLogger(__name__)
