#!/usr/bin/env python
# Copyright (c) 2013 Cloudera, Inc. All rights reserved.
"""
Utility to create Hive Metastore Database tables
"""
import logging
import sys
import os
import re
from subprocess import call, Popen, PIPE

class CreateHiveTablesHandler:
  LEGAL_DBS = ("mysql", "oracle", "postgres")
  SCRIPTS_DIR_REL_PATH = "scripts/metastore/upgrade/"
  """
  regex for DDL file names
  group 0 is the whole string
  group 1 is the version
  group 2 is the first version part
  group 3 is the second version part
  group 4 is the third verison part
  """
  DDL_REGEX = re.compile("hive-schema-((\d+)\.(\d+)\.(\d+))\.[a-z]+\.sql")
  DDL_FORMAT_STR = "hive-schema-%s.%s.sql"
  def __init__(self, hive_home, dbtype, dbname, user, properties_path):
    self.hive_home = hive_home
    self.dbname = dbname
    self.user = user
    self.properties_path = properties_path
    if dbtype == "postgresql":
      # convert to what hive directories expect
      dbtype = "postgres"
    self.dbtype = dbtype
    if dbtype not in CreateHiveTablesHandler.LEGAL_DBS:
      raise Exception("Illegal dbtype: %s. Must be one of %s" % \
                      (dbtype, CreateHiveTablesHandler.LEGAL_DBS))

  def _detect_version(self):
    """
    Return string hive version, such as "0.4.1" or "0.10.0". Detects by
    checking the ddl file with the highest version number in the mysql
    directory of hive lib. This is always present in CDH3+, unlike
    postgres or oracle.
    """
    results = self._list_ddl_in_lib("mysql")
    results = self._sort_ddl_files(results)
    if len(results) == 0:
      raise Exception("Failed to detect version. No ddl files in mysql lib dir")
    return self._get_version_from_ddl_filename(results[0])
  
  def _list_ddl_in_lib(self, dbtype):
    return self._list_ddls_in_dir(
      self._get_lib_dir(dbtype))
  
  def _get_lib_dir(self, dbtype=None):
    """
    Get hive lib directory with the given dbType.
    If dbType is None or omitted, uses self.dbType
    """
    if None == dbtype:
      dbtype = self.dbtype
    return os.path.join(
        self.hive_home,
        CreateHiveTablesHandler.SCRIPTS_DIR_REL_PATH,
        dbtype)
  
  def _list_ddls_in_dir(self, dir):
    if not os.path.exists(dir):
      logging.info("Path does not exist: %s" % (dir,))
      return []
    logging.info("listing ddls in dir: %s" % (dir,))
    return [f for f in os.listdir(dir) if CreateHiveTablesHandler.DDL_REGEX.match(f)]
  
  def _sort_ddl_files(self, ddls):
    """
    Sort ddls in descending version order
    """
    def compare_ddl_files(x, y):
      """
      compare two ddl files by version number
      """
      xMatch = CreateHiveTablesHandler.DDL_REGEX.match(x)
      yMatch = CreateHiveTablesHandler.DDL_REGEX.match(y)
      if not xMatch:
        raise Exception("Not a ddl file: %s" % (x,))
      if not yMatch:
        raise Exception("Not a ddl file: %s" % (y,)) 
      
      # compare versions by most significant digit first
      for groupNum in (2, 3, 4):
        xValue = int(xMatch.group(groupNum))
        yValue = int(yMatch.group(groupNum))
        if xValue != yValue:
          return yValue - xValue

      return 0    
    
    return sorted(ddls, cmp=compare_ddl_files)
    
  def _get_version_from_ddl_filename(self, f):
    match = CreateHiveTablesHandler.DDL_REGEX.match(f)
    if match:
      return match.group(1)
    else:
      raise Exception("Not a ddl file: %s" % (f,))

  def _get_cm_ddl_path(self):
    return os.path.join(os.path.dirname(__file__), 'ddl')

  def _get_ddl(self, version):
    ret = self._get_ddl_from_dir(self._get_lib_dir(), version) or \
      self._get_ddl_from_dir(self._get_cm_ddl_path(), version)
    if None == ret:
      raise Exception("Could not find ddl script for hive version %s" % (version,))
    return ret;
  
  def _get_ddl_from_dir(self, dir, version):
    path = os.path.join(
      dir,
      CreateHiveTablesHandler.DDL_FORMAT_STR % (version, self.dbtype))
    if not os.path.exists(path):
      logging.info("No ddl file at location: %s" % (path,))
      return None
    return path
  
  def _run_sql(self, count_query=None, sqlfile=None):
    """
    Execute a count query or the contents of a sqlfile, and return
    the stdout
    """
    flag = None
    if count_query:
      flag = "--countquery"
    elif sqlfile:
      flag = "--file"
    else:
      raise Exception("must provide count_query or sqlfile")
    
    classpath = ""
    CMF_SERVER_ROOT = "CMF_SERVER_ROOT"
    if CMF_SERVER_ROOT in os.environ:
      classpath = os.path.join(os.environ[CMF_SERVER_ROOT], "lib", "*")
    else:
      classpath = "/usr/share/cmf/lib/*"
    for jarvar in ["CLOUDERA_MYSQL_CONNECTOR_JAR", "CLOUDERA_POSTGRESQL_JDBC_JAR", "CLOUDERA_ORACLE_CONNECTOR_JAR"]:
      if jarvar in os.environ:
        classpath += ":" + os.environ[jarvar]
    logging.info("classpath=%s" % (classpath,))
    java = os.path.join(os.environ["JAVA_HOME"], "bin", "java")
    logging.info("java=%s" % (java,))
    process = Popen([java,
        "-Djava.net.preferIPv4Stack=true", # Disable IPv6.
        "-cp",
        classpath,
        "com.cloudera.enterprise.dbutil.SqlRunner", self.properties_path,
        flag, count_query or sqlfile],
      stdout=PIPE)
    stdout, stderr = process.communicate()
    print stdout
    if process.returncode != 0:
      raise Exception("Problem running sql. Aborting.")
    return stdout
  
  def _parse_count(self, stdout):
    """
    Parse a count from stdout. Assumes first line that is an int is the count,
    ignoring whitespace.
    """
    count = None
    for line in stdout.splitlines():
      try:
        count = int(line.strip())
        break
      except Exception:
        pass # not an int
    
    if None == count:
      print stdout # it was redirected, never printed before
      raise Exception("Error: Problem querying mysql table count. Couldn't identify count in output")
    
    logging.info("Found %s tables" % (count,))
    return count
  
  def _has_tables(self, query):
    stdout = self._run_sql(count_query=query)
    count = self._parse_count(stdout)
    return count > 0
  
  def create_tables(self):
    version = self._detect_version()
    logging.info("Detected Hive Version: %s" % (version,))
    ddl = self._get_ddl(version)
    logging.info("Using ddl file: %s" % (ddl,))
    
    queries = {
      "mysql":"select count(*) from INFORMATION_SCHEMA.tables where table_schema = '%s'" % (self.dbname),
      "oracle":"select count(*) from user_tables",
      "postgres":"select count(*) from pg_tables where tableowner = '%s'" % (self.user,)
    }
    
    if self._has_tables(queries[self.dbtype]):
      logging.info("Hive Metastore Database Tables already exist. Skipping.")
      return
    self._run_sql(sqlfile=ddl)

def parse_properties(path):
  ret = {}
  f = open(path, 'r')
  try:
    for line in f:
      line = line[:-1] # all of these end in a newline, strip it off
      if len(line) == 0 or line[0] == "#":
        continue
      # split at most once, so "=" can be in the password or something
      parts = line.split("=", 1)
      if len(parts) != 2:
        raise Exception("Malformed properties file: %s. Bad line: '%s'" % (path, repr(line)))
      ret[parts[0]] = parts[1]
  finally:
    f.close()
  
  return ret     

if __name__ == '__main__':
  logging.basicConfig(level=logging.INFO)
  if len(sys.argv) != 3:
    raise Exception("usage: create_metastore_tables.py hive_home /path/to/metastore_db_py.properties")
  hiveHome = sys.argv[1]
  properties_path = sys.argv[2]
  
  properties = parse_properties(properties_path)
  dbtype = properties["dbtype"]
  host = properties["host"]
  port = properties["port"]
  dbname = properties["dbname"]
  user = properties["user"]
  
  logging.info("Creating Hive Metastore Tables using " +
    "hive_home=%s, dbtype=%s, host=%s, port=%s, dbname=%s, user=%s" % (hiveHome, dbtype, host, port, dbname, user))
  
  CreateHiveTablesHandler(hiveHome, dbtype, dbname, user, properties_path).create_tables()
