// (c) Copyright 2013 Cloudera, Inc. All rights reserved.
define([
  'knockout',
  'underscore',
  'cloudera/Util',
  // Past this line we don't need the reference.
  'cloudera/knockout/ko.emphasisText'
], function(ko, _, Util) {

  // We don't want the user selecting any of the metrics that belong to any
  // of these entities.
  var IGNORE_ENTITIES = ['ACTIVITY', 'ATTEMPT', 'CLUSTER'];

  // Function that is "bound" to a metric instance. Returns true if this metric
  // could conceivably be concerned with the given term.
  var matches = function(term) {
    if (!term) {
      return true;
    }
    // Do a case-insensitive match.
    var matcher = new RegExp(term, 'i');
    return matcher.test(this.name)
      || matcher.test(this.displayName)
      || matcher.test(this.description);
  };

  var Entity = function(entity, name) {
    var self = this;

    // Create the list of metrics.
    var metrics = {};
    _.each(entity, function(metricsContainer, versionName) {
      _.each(metricsContainer, function(metric, metricShortName) {
        if (!metrics[metricShortName]) {
          var versions = [];
          // A versionName of "enterprise" effectively means that this metric
          // is not versioned.
          if (versionName !== 'enterprise') {
            versions.push(versionName);
          }
          metrics[metricShortName] = {
            name: metricShortName,
            displayName: metric.name,
            description: metric.description,
            versions: versions,
            matches: matches
          };
        } else {
          // Update the versions attribute of the metric.
          metrics[metricShortName].versions.push(versionName);
        }
      });
    });

    self.name = name;
    self.expanded = ko.observable(false);
    self.metrics = _.sortBy(metrics, 'displayName');

    self.toggleExpansion = function() {
      self.expanded(!self.expanded());
    };
  };

  // options:
  // - metrics: Metrics generated by MetricMapsContainer as JSON.
  // - metricListSelector: CSS selector of element that will contain list.
  // - dialogId: ID of the dialog for automatic dismissal when metric is selected.
  var SelectMetricDialog = function(options) {
    var self = this;

    self.metrics = ko.observableArray();
    self.entities = ko.observableArray();
    self.filterText = ko.observable();

    // Provide a list of entities that contains metrics that match the
    // filterText.
    self.filteredEntities = ko.computed(function() {
      var filterText = self.filterText();
      if (!filterText) {
        return self.entities();
      } else {
        filterText = filterText.toLowerCase();
        return _.filter(self.entities(), function(entity) {
          return _.find(entity.metrics, function(metric) {
            return metric.matches(filterText);
          });
        });
      }
    });

    self.createEntity = function(entityFromServer, entityName) {
      if (IGNORE_ENTITIES.indexOf(entityName) !== -1) {
        return;
      }
      var entity = new Entity(entityFromServer, entityName);
      // Monkeypatch the entity instance with a filtered metrics property.
      // We can't do this in the Entity ctor because it doesn't have access
      // to the filterText property.
      entity.filteredMetrics = ko.computed(function() {
        return _.filter(entity.metrics, function(metric) {
          return metric.matches(self.filterText());
        });
      });
      return entity;
    };

    // The server is expected to return data in this format:
    // values:
    // - ALL_CAPS_ENTITY:
    //   - cdh_version:
    //     - metricShortName:
    //       - name: "Name"
    //       - description: "Description"
    // This method updates the list of entities. Entities have a name
    // and a list of associated metrics (as created by createMetric).
    // These metrics are sorted alphabetically by display name.
    // Metrics are de-duped by metricShortName.
    self.createEntities = function(data) {
      // Because the _.map can return null values for entities we skip, we use
      // compact to remove the falsy values from the list.
      var entities = _.compact(
        _.map(data, self.createEntity));
      self.entities(_.sortBy(entities, 'name'));
    };

    // Called when one of the metrics is clicked in the dialog. The intent
    // is that the chart search page is subscribed to this event and will add
    // this metric to the appropriate search box.
    self.addMetric = function(metric) {
      $.publish('chartSearch.addMetric', [metric.name]);
      $('#' + options.dialogId).modal('hide');
    };

    // Populate the view model with the metric values.
    self.createEntities(options.metrics.values);
    // Bind this view model to the metric list selector.
    ko.applyBindings(self, $(options.metricListSelector)[0]);
  };

  return SelectMetricDialog;
});