// (c) Copyright 2012 Cloudera, Inc. All rights reserved.
define([
  "cloudera/Util",
  "cloudera/common/I18n",
  "cloudera/common/Url",
  "underscore",
  "knockout",
  "komapping"
], function (Util, I18n, Url, _, ko, komapping) {

  function generateHtmlList(list) {
    var html = _.map(list, function(item){
      // Need a container because $.html returns
      // the innerHtml.
      var $container = $("<ul>");
      var $li = $("<li>").text(item);
      $container.append($li);
      return $container.html();
    });
    return "<ul>" + html.join("") + "</ul>";
  }

  var viewModel;
  var State = function(data, parent) {
    var self = this;
    komapping.fromJS(data, {}, self);

    self.product = parent.product;
    self.version = parent.version;
    self.isUpgradeNeeded = parent.isUpgradeNeeded;

    if (_.isUndefined(self.errors)) {
      self.errors = ko.observableArray();
    }

    if (_.isUndefined(self.warnings)) {
      self.warnings = ko.observableArray();
    }

    self.showErrors = function(vm, event) {
      $.publish("showAlert", [self.errorsHtml(), self.errorsTitle()]);
      if (event) {
        event.preventDefault();
      }
    };

    self.showWarnings = function(vm, event) {
      $.publish("showAlert", [self.warningsHtml(), self.warningsTitle()]);
      if (event) {
        event.preventDefault();
      }
    };

    self.errorsHtml = ko.computed(function() {
      return generateHtmlList(self.errors());
    });

    self.warningsHtml = ko.computed(function() {
      return generateHtmlList(self.warnings());
    });

    self.errorsTitle = ko.computed(function() {
      return "<i class='IconError16x16'></i> " + I18n.t("ui.nErrors", self.errors().length);
    });

    self.warningsTitle = ko.computed(function() {
      return "<i class='IconWarning16x16'></i> " + I18n.t("ui.nWarnings", self.warnings().length);
    });

    /**
     * Downloads a set of parcels (with the specific product/version)
     * from the remote server.
     */
    self.download = function() {
      if (self.isRemotelyAvailable()) {
        var callback = function() {
          self.stage('DOWNLOADING');
          self.progress(0);
          self.totalProgress(0);
          self.count(0);
          self.totalCount(0);
        };
        self.execute(parent.urls.downloadUrl(), callback);
      }
    };

    /**
     * Deletes the specific parcels from the local repository.
     */
    self.deleteFromLocal = function() {
      if (self.isDownloaded()) {
        var callback = function() {
          self.stage('AVAILABLE_REMOTELY');
          self.progress(0);
          self.totalProgress(0);
          self.count(0);
          self.totalCount(0);
        };
        self.execute(parent.urls.deleteUrl(), callback);
      }
    };

    /**
     * Distributes the specific parcels to all the machines
     * in this cluster.
     */
    self.distribute = function() {
      if (self.isDownloaded()) {
        var callback = function() {
          self.stage('DISTRIBUTING');
          self.progress(0);
          self.totalProgress(0);
          self.count(0);
          self.totalCount(0);
        };
        self.execute(parent.urls.distributeUrl(), callback);
      }
    };

    /**
     * Reverts the distribute process by removing the
     * specific parcels from all the machines in this cluster.
     */
    self.undistribute = function() {
      if (self.isDistributed()) {
        self.doUndistribute();
      }
    };

    /**
     * The helper funciton that is used to actually perform
     * the undistribute.
     **/
    self.doUndistribute = function() {
      var callback = function() {
        self.stage('UNDISTRIBUTING');
        self.progress(0);
        self.totalProgress(0);
        self.count(0);
        self.totalCount(0);
      };
      self.execute(parent.urls.undistributeUrl(), callback);
    };

    /**
     * Undistribute the parcel if we are currently in the
     * upgrade mode. This is needed because clicking the upgrade
     * button activates the parcel then performs the upgrade. If
     * a user cancels out of the upgrade popup we need to deactivate
     * before we can undistribute the parcel.
     **/
    self.undistributeWithUpgrade = function() {
      if (self.isClusterUpgradeRequired()) {
        if (self.stage() === "ACTIVATED") {
          self.execute(parent.urls.deactivateUrl(), function() {
            self.doUndistribute();
          });
        }else {
          self.doUndistribute();
        }
      }
    };

    /**
     * Initiates a cluster upgrade. THis involves first
     * activating the parcel, then running then launching
     * the upgrade popup.
     */
    self.upgrade = function() {
      if (self.isClusterUpgradeRequired()) {
        if (viewModel.loader) {
          var $link = viewModel.loader.ajaxPopupLink();
          var url = new Url(parent.urls.upgradeUrl());
          var params = url.getParamsObject();
          params.parcelVersion = self.version();
          params.parcelProduct = self.product();
          url.setParamsObject(params);
          $link.attr('href', url.getHref());
          $link.trigger("click");
        }
      }
    };

    /**
     * Shows the upgrade popup.
     **/
    self.showRestartPopup = function(isActivate) {
      if (viewModel.loader) {
        var $link = viewModel.loader.ajaxPopupLink();
        var url = new Url(parent.urls.restartPopupUrl());
        var params = url.getParamsObject();
        params.isActivate = isActivate;
        url.setParamsObject(params);
        $link.attr('href', url.getHref());
        $link.trigger("click");
      }
    };

    /**
     * Activates the specific product/version for this cluster.
     */
    self.activate = function() {
      if (self.isDistributed()) {
        var callback = function() {
          self.stage('ACTIVATED');
          self.progress(0);
          self.totalProgress(0);
          self.count(0);
          self.totalCount(0);
          self.showRestartPopup(true);
        };
        self.execute(parent.urls.activateUrl(), callback);
      }
    };

    /**
     * Deactivates the specific product/version for this cluster.
     * The system default packages will be used instead.
     */
    self.deactivate = function() {
      if (self.isActivated()) {
        var callback = function() {
          self.stage('DISTRIBUTED');
          self.progress(0);
          self.totalProgress(0);
          self.count(0);
          self.totalCount(0);
          self.showRestartPopup(false);
        };
        self.execute(parent.urls.deactivateUrl(), callback);
      }
    };

    self.cancel = function() {
      self.execute(parent.urls.cancelUrl());
    };

    /**
     * Executes a server call to a specific URL.
     */
    self.execute = function(url, callback) {
      var queryParams = { product: self.product(),
        version: self.version()
      };
      $.post(url, queryParams, function(response) {
        var jsonResponse = Util.filterJsonResponseError(response);
        if (jsonResponse.message !== "OK") {
          $.publish('showError', [jsonResponse.message]);
        } else if ($.isFunction(callback)) {
          callback();
        }
        // After the execution response, expedite the update.
        if (viewModel.loader) {
          viewModel.loader.update();
        }
      });
    };

    /**
     * @return true to display all the panels when the active flag is unchecked,
     * or true to display only the activated panels.
     */
    self.showPanel = ko.computed(function() {
      return self.isActivated() || !(viewModel.activeOnly());
    });

    /**
     * @return true to display the progress stage.
     */
    self.showStage = ko.computed(function() {
      if (self.isActivated() || self.isDistributing() || self.isDownloading() || self.isUndistributing()) {
        return false;
      } else {
        return true;
      }
    });

    /**
     * @return true if the specific parcels are in the remotely available stage.
     */
    self.isRemotelyAvailable = ko.computed(function() {
      return self.stage() === 'AVAILABLE_REMOTELY';
    });

    /**
     * @return true if the specific parcels are in the downloaded stage.
     */
    self.isDownloaded = ko.computed(function() {
      return self.stage() === 'DOWNLOADED';
    });

    /**
     * @return true if the specific parcels are in the activated stage.
     */
    self.isActivated = ko.computed(function() {
      return (self.stage() === 'ACTIVATED' && !self.isClusterUpgradeRequired());
    });

    /**
     * @return true if the specific parcels are in the distributed stage.
     */
    self.isDistributed = ko.computed(function() {
      return (self.stage() === 'DISTRIBUTED' && !self.isClusterUpgradeRequired());
    });

    /**
     * @return true if the specific parcels are still getting downloaded.
     */
    self.isDownloading = ko.computed(function() {
      return self.stage() === 'DOWNLOADING';
    });

    /**
     * @return true if the specific parcels are still getting distributed/unpacked.
     */
    self.isDistributing = ko.computed(function() {
      return self.stage() === 'DISTRIBUTING';
    });

    /**
     * @return true if the specific parcels are getting removed from the agents.
     */
    self.isUndistributing = ko.computed(function() {
      return self.stage() === 'UNDISTRIBUTING';
    });

    /**
     * @return true if we need to perform a cluster upgrade.
     */
    self.isClusterUpgradeRequired = ko.computed(function() {
      var isUpgradeNeeded = false;
      // This is an optional flag. Most of the tests
      // do not have this flag set in the JSON.
      if (_.isFunction(self.isUpgradeNeeded)) {
        isUpgradeNeeded = self.isUpgradeNeeded();
      }
      return (((self.stage() === 'DISTRIBUTED') || (self.stage() === 'ACTIVATED'))
              && isUpgradeNeeded);
    });

    /**
     * @return "progress / totalProgress" as a string.
     */
    self.progressSummary = ko.computed(function() {
      if (self.progress() > 0 || self.totalProgress() > 0) {
        return self.progress() + " / " + self.totalProgress();
      } else {
        return "";
      }
    });

    /**
     * @return "count / totalCount" as a string.
     */
    self.countSummary = ko.computed(function() {
      if (self.count() > 0 || self.totalCount() > 0) {
        return self.count() + " / " + self.totalCount();
      } else {
        return "";
      }
    });

    /**
     * @return count / totalCount as a percentage string.
     */
    self.countPct = ko.computed(function() {
      if (self.totalCount() > 0) {
        var val = self.count() * 100 / self.totalCount();
        return val.toFixed(0) + "%";
      } else {
        return "0%";
      }
    });

    /**
     * @return progress / totalProgress as a percentage string.
     */
    self.progressPct = ko.computed(function() {
      if (self.totalProgress() > 0) {
        var val = self.progress() * 100 / self.totalProgress();
        return val.toFixed(0) + "%";
      } else {
        return "0%";
      }
    });

    /**
     * @return combined progress / combined total progress as a percentage string.
     */
    self.combinedProgressPct = ko.computed(function() {
      if (self.combinedTotalProgress() > 0) {
        var val = self.combinedProgress() * 100 / self.combinedTotalProgress();
        return val.toFixed(0) + "%";
      } else {
        return "0%";
      }
    });
  };

  /**
   * Used by knockout mappings to create a viewModel from a JavaScript object
   * automatically. Override the object creation process that represent State.
   */
  var mapping = {
    'state': {
      create: function(options) {
        return new State(options.data, options.parent);
      }
    }
  };

  /**
   * Initial state.
   * Note, this viewModel is created outside the Parcels' function scope,
   * so it is shared across multiple Parcels instances.
   * So there can only be one instance of this Parcels object on a page.
   */
  viewModel = komapping.fromJS({
    isLoading: false,
    activeOnly: ko.observable(false),
    data: []
  }, mapping);

  /**
   * options {
   *   container:    (required) "selector of the container DOM object",
   *   updateUri:    (required) "the uri for fetching the JSON",
   *   updateParams: (optional) "the parameters for fetching the JSON",
   *   afterUpdate:  (optional) "a callback to execute after each update"
   * }
   */
  var Parcels = function(options) {

    var self = this, $container = $(options.container);

    /**
     * @return true when user is interacting with the dropdown.
     */
    self.isInteracting = function() {
      return $container.find(".actionsMenu.open").length > 0;
    };

    /**
     * @return the selector for the ajax popup link.
     */
    self.ajaxPopupLink = function() {
      return $container.find(".ajax-popup-link");
    };

    /**
     * Group by active/non active state.
     */
    self.groupByProduct = function(data) {
      var result = [];
      _.each(data, function(d) {
        // Group the states by product/version.
        var statesByProduct = _.groupBy(d.states, function(s) {
          return s.product;
        });

        var statesByProductArr = [];
        _.each(statesByProduct, function(states, product) {
          _.each(states, function(pvState) {
            // This is some what hacky that the State object,
            // needs to know about the urls that performs the various actions.
            // But the mapping create method only exposes the arguments:
            // options.data and options.parent.
            //
            // This adds the urls property to the pvState object, which is the parent object
            // of a State object.
            pvState.urls = d.cluster.urls;
          });
          statesByProductArr.push(_.sortBy(states, function(s) {
            return s.version;
          }).reverse());
        });

        // result looks like
        // [ {
        //   cluster: {
        //     name: "...",
        //   },
        //   statesByProduct: [
        //     [ {
        //       urls: ...
        //       product: 'CDH',
        //       version: '1.0',
        //       state: {
        //             stage: "AVAILABLE_REMOTELY",
        //             count: ...'
        //       }
        //     }, {
        //       // another version
        //     }], [
        //       // another product
        //     ]
        // }, {
        //   // another cluster
        // } ]
        // We cannot combine product, version and members of the state into a single object
        // because knockout mapping requires a name when overriding the constructor.
        result.push({
          cluster: d.cluster,
          statesByProduct: statesByProductArr
        });
      });
      return result;
    };

    /**
     * Updates the status.
     */
    self.update = function() {

      viewModel.isLoading(true);
      $.post(options.updateUri, options.updateParams || {}, function(data) {
        if (!self.isInteracting()) {
          komapping.fromJS({
            isLoading: false,
            activeOnly: ko.observable(viewModel.activeOnly()),
            data: self.groupByProduct(data)
          }, viewModel);
          if (_.isFunction(options.afterUpdate)) {
            options.afterUpdate();
          }
        } else {
          // Do not update the view model
          // because user is interacting with the UI.
          viewModel.isLoading(false);
        }
      }, "json");
    };

    self.start = function() {
      self.update();
      if (!Util.getTestMode()) {
        self._intervalId = window.setInterval(_.bind(self.update, self), 5000);
      }
    };

    self.stop = function() {
      if (!Util.getTestMode()) {
        if (self._intervalId) {
          window.clearInterval(self._intervalId);
          self._intervalId = undefined;
        }
      }
    };

    // Expose for testing purposes.
    self.viewModel = viewModel;

    // The viewModel needs a reference to
    // This object, so it can expedite the update
    // process when an action is performed.
    viewModel.loader = self;


  };

  return Parcels;
});
